#include <math.h>

#include "algebra.h"
#include "os.h"


// First some misc functions used in matrix inversion

// Matrix:
// | a b |
// | c d |
static	inline float det2x2( const float a,const float  b,const float  c,const float  d) {
	return  a * d - b * c;
}

// Matrix:
// | a1 a2 a3 |
// | b1 b2 b3 |
// | c1 c2 c3 |
static	inline float det3x3(const float  a1,const float  a2,const float  a3,const float  b1,const float  b2,const float  b3,const float  c1,const float  c2,const float  c3 ) {
    return	  a1 * det2x2( b2, b3, c2, c3 )
			- b1 * det2x2( a2, a3, c2, c3 )
			+ c1 * det2x2( a2, a3, b2, b3 );
}

// Matrix:
// | a1 a2 a3 a4 |
// | b1 b2 b3 b4 |
// | c1 c2 c3 c4 |
// | d1 d2 d3 d4 |
static	inline float det4x4( const float *m ) {
    float a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4;

	a1 = m[element(0,0)]; b1 = m[element(0,1)]; 
	c1 = m[element(0,2)]; d1 = m[element(0,3)];

	a2 = m[element(1,0)]; b2 = m[element(1,1)]; 
	c2 = m[element(1,2)]; d2 = m[element(1,3)];

	a3 = m[element(2,0)]; b3 = m[element(2,1)]; 
	c3 = m[element(2,2)]; d3 = m[element(2,3)];

	a4 = m[element(3,0)]; b4 = m[element(3,1)]; 
	c4 = m[element(3,2)]; d4 = m[element(3,3)];

    return	  a1 * det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4)
			- b1 * det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4)
			+ c1 * det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4)
			- d1 * det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);
}

// Construct the adjoint matrix
// | a1 a2 a3 a4 |
// | b1 b2 b3 b4 |
// | c1 c2 c3 c4 |
// | d1 d2 d3 d4 |

static	inline void adjoint( float *out,const float *in) {
    float a1, a2, a3, a4, b1, b2, b3, b4;
    float c1, c2, c3, c4, d1, d2, d3, d4;


    /* assign to individual variable names to aid  */
    /* selecting correct values  */

	a1 = in[element(0,0)]; b1 = in[element(0,1)]; 
	c1 = in[element(0,2)]; d1 = in[element(0,3)];

	a2 = in[element(1,0)]; b2 = in[element(1,1)]; 
	c2 = in[element(1,2)]; d2 = in[element(1,3)];

	a3 = in[element(2,0)]; b3 = in[element(2,1)];
	c3 = in[element(2,2)]; d3 = in[element(2,3)];

	a4 = in[element(3,0)]; b4 = in[element(3,1)]; 
	c4 = in[element(3,2)]; d4 = in[element(3,3)];


    /* row column labeling reversed since we transpose rows & columns */

    out[element(0,0)]  =   det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4);
    out[element(1,0)]  = - det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4);
    out[element(2,0)]  =   det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4);
    out[element(3,0)]  = - det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);
        
    out[element(0,1)]  = - det3x3( b1, b3, b4, c1, c3, c4, d1, d3, d4);
    out[element(1,1)]  =   det3x3( a1, a3, a4, c1, c3, c4, d1, d3, d4);
    out[element(2,1)]  = - det3x3( a1, a3, a4, b1, b3, b4, d1, d3, d4);
    out[element(3,1)]  =   det3x3( a1, a3, a4, b1, b3, b4, c1, c3, c4);
        
    out[element(0,2)]  =   det3x3( b1, b2, b4, c1, c2, c4, d1, d2, d4);
    out[element(1,2)]  = - det3x3( a1, a2, a4, c1, c2, c4, d1, d2, d4);
    out[element(2,2)]  =   det3x3( a1, a2, a4, b1, b2, b4, d1, d2, d4);
    out[element(3,2)]  = - det3x3( a1, a2, a4, b1, b2, b4, c1, c2, c4);
        
    out[element(0,3)]  = - det3x3( b1, b2, b3, c1, c2, c3, d1, d2, d3);
    out[element(1,3)]  =   det3x3( a1, a2, a3, c1, c2, c3, d1, d2, d3);
    out[element(2,3)]  = - det3x3( a1, a2, a3, b1, b2, b3, d1, d2, d3);
    out[element(3,3)]  =   det3x3( a1, a2, a3, b1, b2, b3, c1, c2, c3);
}

void	skewsymm(float *dest,const float *src) {
	identitym(dest);

	dest[element(0,0)]	=	0;
	dest[element(1,1)]	=	0;
	dest[element(2,2)]	=	0;
	dest[element(0,1)]	=	-src[COMP_Z];
	dest[element(0,2)]	=	src[COMP_Y];
	dest[element(1,2)]	=	-src[COMP_X];
	dest[element(1,0)]	=	src[COMP_Z];
	dest[element(2,0)]	=	-src[COMP_Y];
	dest[element(2,1)]	=	src[COMP_X];
}


// Actually invert the matrix s to d
int			invertm(float *d,const float *s)	{
	int		i;
    float det;

    adjoint( d, s );

    det = det4x4( s );

    if ( fabs( det ) < 1e-5) {				// FIXME: 1e-5 doesn't seem right here
		identitym(d);
        return 1;
    }

    for (i=0; i<16; i++)
		d[i]		=	d[i] / det;

	return 0;
}

// Take the determinant of the matrix
float	determinantm(const float *r) {
	return det4x4(r);
}

// Construct the identity matrix
void	identitym(float *r) {
	int i;

	for (i=0;i<16;i++) r[i]	=	0;

	r[element(0,0)]	=	1;
	r[element(1,1)]	=	1;
	r[element(2,2)]	=	1;
	r[element(3,3)]	=	1;
}

// Construct the translate matrix
void	translatem(float *r,const float tx,const float ty,const float tz) {
	identitym(r);

	r[element(0,3)]	=	tx;
	r[element(1,3)]	=	ty;
	r[element(2,3)]	=	tz;
}

// Construct the scale matrix
void	scalem(float *r,const float sx,const float sy,const float sz) {
	identitym(r);

	r[element(0,0)]	=	sx;
	r[element(1,1)]	=	sy;
	r[element(2,2)]	=	sz;
}

// Construct the rotate matrix
void	rotatem(float *r,const float *v,const float angle) {
	rotatem(r,v[COMP_X],v[COMP_Y],v[COMP_Z],angle);
}

// Construct the rotate matrix
void	rotatem(float *r,float x,float y,float z,const float angle) {
	float	s	= (float) cos(radians(angle/2));
	float	sp	= (float) sin(radians(angle/2));
	float	l	=	(float) sqrt(x*x + y*y + z*z);
	x			/=	l;
	y			/=	l;
	z			/=	l;
	float	a	= x*sp;
	float	b	= y*sp;
	float	c	= z*sp;
	l			=	(float) sqrt(a*a + b*b + c*c + s*s);
	a			/=	l;
	b			/=	l;
	c			/=	l;
	s			/=	l;

	r[element(0,0)] = 1-2*b*b-2*c*c;
	r[element(0,1)] = 2*a*b-2*s*c;
	r[element(0,2)] = 2*a*c+2*s*b;
	r[element(0,3)] = 0;
	r[element(1,0)] = 2*a*b+2*s*c;
	r[element(1,1)] = 1-2*a*a-2*c*c;
	r[element(1,2)] = 2*b*c-2*s*a;
	r[element(1,3)] = 0;
	r[element(2,0)] = 2*a*c-2*s*b;
	r[element(2,1)] = 2*b*c+2*s*a;
	r[element(2,2)] = 1-2*a*a-2*b*b;
	r[element(2,3)] = 0;
	r[element(3,0)] = 0;
	r[element(3,1)] = 0;
	r[element(3,2)] = 0;
	r[element(3,3)] = 1;
}

void	skewm(float *r,const float angle,const float dx1,const float dy1,const float dz1,const float dx2,const float dy2,const float dz2) {
	vector	v1,v2;
	float	tantheta	=	(float) tan(angle*PI/(double) 180);

	v1[COMP_X]	=	dx1;
	v1[COMP_Y]	=	dy1;
	v1[COMP_Z]	=	dz1;
	normalizev(v1);
	
	v2[COMP_X]	=	dx2;
	v2[COMP_Y]	=	dy2;
	v2[COMP_Z]	=	dz2;
	normalizev(v2);

	r[element(0,0)]	=	1+v1[COMP_X]*tantheta*v2[COMP_X];
	r[element(0,1)]	=	v1[COMP_Y]*tantheta*v2[COMP_X];
	r[element(0,2)]	=	v1[COMP_Z]*tantheta+v2[COMP_X];
	r[element(0,3)]	=	0;
	r[element(1,0)]	=	v1[COMP_X]*tantheta*v2[COMP_Y];
	r[element(1,1)]	=	1+v1[COMP_Y]*tantheta*v2[COMP_Y];
	r[element(1,2)]	=	v1[COMP_Z]*tantheta+v2[COMP_Y];
	r[element(1,3)]	=	0;
	r[element(2,0)]	=	v1[COMP_X]*tantheta*v2[COMP_Z];
	r[element(2,1)]	=	v1[COMP_Y]*tantheta*v2[COMP_Z];
	r[element(2,2)]	=	1+v1[COMP_Z]*tantheta+v2[COMP_Z];
	r[element(2,3)]	=	0;
	r[element(3,0)]	=	0;
	r[element(3,1)]	=	0;
	r[element(3,2)]	=	0;
	r[element(3,3)]	=	1;
}


#define	S		0
#define	VX		1
#define	VY		2
#define	VZ		3

void	rotatem(float *to,const float *from) {
	to[element(0,0)]	=	1-2*from[VY]*from[VY] - 2*from[VZ]*from[VZ];
	to[element(0,1)]	=	2*from[VX]*from[VY] - 2*from[S]*from[VZ];
	to[element(0,2)]	=	2*from[VX]*from[VZ] + 2*from[S]*from[VY];
	to[element(1,0)]	=	2*from[VX]*from[VY] + 2*from[S]*from[VZ];
	to[element(1,1)]	=	1-2*from[VX]*from[VX] - 2*from[VZ]*from[VZ];
	to[element(1,2)]	=	2*from[VY]*from[VZ] - 2*from[S]*from[VX];
	to[element(2,0)]	=	2*from[VX]*from[VZ] - 2*from[S]*from[VY];
	to[element(2,1)]	=	2*from[VY]*from[VZ] + 2*from[S]*from[VX];
	to[element(2,2)]	=	1-2*from[VX]*from[VX] - 2*from[VY]*from[VY];
	to[element(0,3)]	=	0;
	to[element(1,3)]	=	0;
	to[element(2,3)]	=	0;
	to[element(3,3)]	=	1;
	to[element(3,0)]	=	0;
	to[element(3,1)]	=	0;
	to[element(3,2)]	=	0;
}

#undef S
#undef VX
#undef VY
#undef VZ


/////////////////////////////////////////////////////////////////////////////////////////////////////// 
#define SIGN(a, b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define MIN(x,y) ( (x) < (y) ? (x) : (y) )
#define MAX(x,y) ((x)>(y)?(x):(y))


static double PYTHAG(double a, double b)
{
    double at = fabs(a), bt = fabs(b), ct, result;

    if (at > bt)       { ct = bt / at; result = at * sqrt(1.0 + ct * ct); }
    else if (bt > 0.0) { ct = at / bt; result = bt * sqrt(1.0 + ct * ct); }
    else result = 0.0;
    return(result);
}


int dsvd(float **a, int m, int n, float *w, float **v)
{
    int flag, i, its, j, jj, k, l, nm;
    double c, f, h, s, x, y, z;
    double anorm = 0.0, g = 0.0, scale = 0.0;
    double *rv1;
  
    if (m < n) {
        //fprintf(stderr, "#rows must be > #cols \n");
        return(0);
    }
  
    rv1 = (double *) COs::memmalloc((unsigned int) n*sizeof(double));

    /* Householder reduction to bidiagonal form */
    for (i = 0; i < n; i++) 
    {
        /* left-hand reduction */
        l = i + 1;
        rv1[i] = scale * g;
        g = s = scale = 0.0;
        if (i < m) 
        {
            for (k = i; k < m; k++) 
                scale += fabs((double)a[k][i]);
            if (scale) 
            {
                for (k = i; k < m; k++) 
                {
                    a[k][i] = (float)((double)a[k][i]/scale);
                    s += ((double)a[k][i] * (double)a[k][i]);
                }
                f = (double)a[i][i];
                g = -SIGN(sqrt(s), f);
                h = f * g - s;
                a[i][i] = (float)(f - g);
                if (i != n - 1) 
                {
                    for (j = l; j < n; j++) 
                    {
                        for (s = 0.0, k = i; k < m; k++) 
                            s += ((double)a[k][i] * (double)a[k][j]);
                        f = s / h;
                        for (k = i; k < m; k++) 
                            a[k][j] += (float)(f * (double)a[k][i]);
                    }
                }
                for (k = i; k < m; k++) 
                    a[k][i] = (float)((double)a[k][i]*scale);
            }
        }
        w[i] = (float)(scale * g);
    
        /* right-hand reduction */
        g = s = scale = 0.0;
        if (i < m && i != n - 1) 
        {
            for (k = l; k < n; k++) 
                scale += fabs((double)a[i][k]);
            if (scale) 
            {
                for (k = l; k < n; k++) 
                {
                    a[i][k] = (float)((double)a[i][k]/scale);
                    s += ((double)a[i][k] * (double)a[i][k]);
                }
                f = (double)a[i][l];
                g = -SIGN(sqrt(s), f);
                h = f * g - s;
                a[i][l] = (float)(f - g);
                for (k = l; k < n; k++) 
                    rv1[k] = (double)a[i][k] / h;
                if (i != m - 1) 
                {
                    for (j = l; j < m; j++) 
                    {
                        for (s = 0.0, k = l; k < n; k++) 
                            s += ((double)a[j][k] * (double)a[i][k]);
                        for (k = l; k < n; k++) 
                            a[j][k] += (float)(s * rv1[k]);
                    }
                }
                for (k = l; k < n; k++) 
                    a[i][k] = (float)((double)a[i][k]*scale);
            }
        }
        anorm = MAX(anorm, (fabs((double)w[i]) + fabs(rv1[i])));
    }
  
    /* accumulate the right-hand transformation */
    for (i = n - 1; i >= 0; i--) 
    {
        if (i < n - 1) 
        {
            if (g) 
            {
                for (j = l; j < n; j++)
                    v[j][i] = (float)(((double)a[i][j] / (double)a[i][l]) / g);
                    /* double division to avoid underflow */
                for (j = l; j < n; j++) 
                {
                    for (s = 0.0, k = l; k < n; k++) 
                        s += ((double)a[i][k] * (double)v[k][j]);
                    for (k = l; k < n; k++) 
                        v[k][j] += (float)(s * (double)v[k][i]);
                }
            }
            for (j = l; j < n; j++) 
                v[i][j] = v[j][i] = 0.0;
        }
        v[i][i] = 1.0;
        g = rv1[i];
        l = i;
    }
  
    /* accumulate the left-hand transformation */
    for (i = n - 1; i >= 0; i--) 
    {
        l = i + 1;
        g = (double)w[i];
        if (i < n - 1) 
            for (j = l; j < n; j++) 
                a[i][j] = 0.0;
        if (g) 
        {
            g = 1.0 / g;
            if (i != n - 1) 
            {
                for (j = l; j < n; j++) 
                {
                    for (s = 0.0, k = l; k < m; k++) 
                        s += ((double)a[k][i] * (double)a[k][j]);
                    f = (s / (double)a[i][i]) * g;
                    for (k = i; k < m; k++) 
                        a[k][j] += (float)(f * (double)a[k][i]);
                }
            }
            for (j = i; j < m; j++) 
                a[j][i] = (float)((double)a[j][i]*g);
        }
        else 
        {
            for (j = i; j < m; j++) 
                a[j][i] = 0.0;
        }
        ++a[i][i];
    }

    /* diagonalize the bidiagonal form */
    for (k = n - 1; k >= 0; k--) 
    {                             /* loop over singular values */
        for (its = 0; its < 30; its++) 
        {                         /* loop over allowed iterations */
            flag = 1;
            for (l = k; l >= 0; l--) 
            {                     /* test for splitting */
                nm = l - 1;
                if (fabs(rv1[l]) + anorm == anorm) 
                {
                    flag = 0;
                    break;
                }
                if (fabs((double)w[nm]) + anorm == anorm) 
                    break;
            }
            if (flag) 
            {
                c = 0.0;
                s = 1.0;
                for (i = l; i <= k; i++) 
                {
                    f = s * rv1[i];
                    if (fabs(f) + anorm != anorm) 
                    {
                        g = (double)w[i];
                        h = PYTHAG(f, g);
                        w[i] = (float)h; 
                        h = 1.0 / h;
                        c = g * h;
                        s = (- f * h);
                        for (j = 0; j < m; j++) 
                        {
                            y = (double)a[j][nm];
                            z = (double)a[j][i];
                            a[j][nm] = (float)(y * c + z * s);
                            a[j][i] = (float)(z * c - y * s);
                        }
                    }
                }
            }
            z = (double)w[k];
            if (l == k) 
            {                  /* convergence */
                if (z < 0.0) 
                {              /* make singular value nonnegative */
                    w[k] = (float)(-z);
                    for (j = 0; j < n; j++) 
                        v[j][k] = (-v[j][k]);
                }
                break;
            }
            if (its >= 30) {
				COs::memfree((void*) rv1);
                //fprintf(stderr, "No convergence after 30,000! iterations \n");
                return(0);
            }
    
            /* shift from bottom 2 x 2 minor */
            x = (double)w[l];
            nm = k - 1;
            y = (double)w[nm];
            g = rv1[nm];
            h = rv1[k];
            f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2.0 * h * y);
            g = PYTHAG(f, 1.0);
            f = ((x - z) * (x + z) + h * ((y / (f + SIGN(g, f))) - h)) / x;
          
            /* next QR transformation */
            c = s = 1.0;
            for (j = l; j <= nm; j++) 
            {
                i = j + 1;
                g = rv1[i];
                y = (double)w[i];
                h = s * g;
                g = c * g;
                z = PYTHAG(f, h);
                rv1[j] = z;
                c = f / z;
                s = h / z;
                f = x * c + g * s;
                g = g * c - x * s;
                h = y * s;
                y = y * c;
                for (jj = 0; jj < n; jj++) 
                {
                    x = (double)v[jj][j];
                    z = (double)v[jj][i];
                    v[jj][j] = (float)(x * c + z * s);
                    v[jj][i] = (float)(z * c - x * s);
                }
                z = PYTHAG(f, h);
                w[j] = (float)z;
                if (z) 
                {
                    z = 1.0 / z;
                    c = f * z;
                    s = h * z;
                }
                f = (c * g) + (s * y);
                x = (c * y) - (s * g);
                for (jj = 0; jj < m; jj++) 
                {
                    y = (double)a[jj][j];
                    z = (double)a[jj][i];
                    a[jj][j] = (float)(y * c + z * s);
                    a[jj][i] = (float)(z * c - y * s);
                }
            }
            rv1[l] = 0.0;
            rv1[k] = f;
            w[k] = (float)x;
        }
    }
	COs::memfree((void*) rv1);
    return(1);
}


void	exponentmv(float *d,const float *s) {
	matrix	I;
	matrix	R;
	matrix	R2;
	float	rot	=	(float) sqrt(dotvv(s,s));		// Rotation amount in radians
	int		i;
	float	a,b;

	a		=	(float) sin(rot);
	b		=	1 - (float) cos(rot);

	identitym(I);
	R[element(0,0)]	=	0;
	R[element(0,1)]	=	-s[2];
	R[element(0,2)]	=	s[1];
	R[element(0,3)]	=	0;
	R[element(1,0)]	=	s[2];
	R[element(1,1)]	=	0;
	R[element(1,2)]	=	-s[0];
	R[element(1,3)]	=	0;
	R[element(2,0)]	=	-s[1];
	R[element(2,1)]	=	s[0];
	R[element(2,2)]	=	0;
	R[element(2,3)]	=	0;
	R[element(3,0)]	=	0;
	R[element(3,1)]	=	0;
	R[element(3,2)]	=	0;
	R[element(3,3)]	=	1;
	mulmm(R2,R,R);

	for (i=0;i<16;i++) {
		d[i]	=	I[i] + a*R[i] + b*R2[i];
	}
}

void	exponentvm(float *d,const float *s) {
	float	trace	=	s[element(0,0)] + s[element(1,1)] + s[element(2,2)];
	matrix	Rt;
	float	theta;
	matrix	R;

	theta			=	(float) acos((trace - 1)*(float) 0.5);
	trace			=	(float) (2*sin(theta));

	transposem(Rt,s);
	subvv(R,s,Rt);

	d[0]			=	-R[element(1,2)]/trace;
	d[1]			=	R[element(0,2)]/trace;
	d[2]			=	-R[element(0,1)]/trace;
}
