///////////////////////////////////////////////////////////////////////
//
//  File				:	algebra,h
//  Description			:	This file defines vector and matrix types and
//							their misc operations
//
//
//	Ver 11/28/2001
//
////////////////////////////////////////////////////////////////////////


#ifndef ALGEBRA_H
#define ALGEBRA_H

// Include the all mighty global header file
#include "global.h"

#include <math.h>
#include <stdlib.h>

// Component indices
const	int		COMP_X						=	0;
const	int		COMP_Y						=	1;
const	int		COMP_Z						=	2;

const	int		COMP_R						=	0;
const	int		COMP_G						=	1;
const	int		COMP_B						=	2;

// Row major matrix element order (compatible with openGL)
#define	element(row,column)	(row+column*4)

typedef float	vector[3];									// an array of 3 floats
typedef float	matrix[16];									// an array of 16 floats

// Initialization functions
inline	void	initv(float *r,const float x,const float y,const float z) {
	r[COMP_X]	=	x;
	r[COMP_Y]	=	y;
	r[COMP_Z]	=	z;
}

inline	void	initc(float *r,const float red,const float green,const float blue) {
	r[COMP_R]	=	red;
	r[COMP_G]	=	green;
	r[COMP_B]	=	blue;
}

// Vector - vector addition
inline	void	addvv(float *result,const float *s1,const float *s2) {
	result[0]	=	s1[0] + s2[0];
	result[1]	=	s1[1] + s2[1];
	result[2]	=	s1[2] + s2[2];
}

inline	void	addvv(float *result,const float *s1) {
	result[0]	+=	s1[0];
	result[1]	+=	s1[1];
	result[2]	+=	s1[2];
}

// Vector - vector subtraction
inline	void	subvv(float *result,const float *s1,const float *s2) {
	result[0]	=	s1[0] - s2[0];
	result[1]	=	s1[1] - s2[1];
	result[2]	=	s1[2] - s2[2];
}

inline	void	subvv(float *result,const float *s1) {
	result[0]	-=	s1[0];
	result[1]	-=	s1[1];
	result[2]	-=	s1[2];
}

// Vector - float multiplication
inline	void	mulvf(float *result,const float *v,const float m) {
	result[0]	=	v[0]*m;
	result[1]	=	v[1]*m;
	result[2]	=	v[2]*m;
}

inline	void	mulvf(float *result,const float m) {
	result[0]	*=	m;
	result[1]	*=	m;
	result[2]	*=	m;
}

// Vector - vector multiplication (component by component)
inline	void	mulvv(float	*result,const float *s1,const float *s2) {
	result[0]	=	s1[0]*s2[0];
	result[1]	=	s1[1]*s2[1];
	result[2]	=	s1[2]*s2[2];
}

inline	void	mulvv(float *result,const float *s1) {
	result[0]	*=	s1[0];
	result[1]	*=	s1[1];
	result[2]	*=	s1[2];
}

// Vector - vector division (component by component)
inline	void	divvv(float *result,const float *s1,const float *s2) {
	result[0]	=	s1[0] / s2[0];
	result[1]	=	s1[1] / s2[1];
	result[2]	=	s1[2] / s2[2];
}

inline	void	divvv(float *result,const float *s1) {
	result[0]	/=	s1[0];
	result[1]	/=	s1[1];
	result[2]	/=	s1[2];
}


// Vector - vector cross product
inline	void	crossvv(float *result,const float *s1,const float *s2) {
	result[COMP_X]	=	(s1[COMP_Y]*s2[COMP_Z] - s1[COMP_Z]*s2[COMP_Y]);
	result[COMP_Y]	=	(s1[COMP_Z]*s2[COMP_X] - s1[COMP_X]*s2[COMP_Z]);
	result[COMP_Z]	=	(s1[COMP_X]*s2[COMP_Y] - s1[COMP_Y]*s2[COMP_X]);
}

// Vector - vector dot product
inline	float	dotvv(const float *s1,const float *s2) {
	return (float) (s1[COMP_X]*s2[COMP_X] + s1[COMP_Y]*s2[COMP_Y] + s1[COMP_Z]*s2[COMP_Z]);
}

// double version, just in case
inline	double	ddotvv(const float *s1,const float *s2) {
	return (double) (s1[COMP_X]*s2[COMP_X] + s1[COMP_Y]*s2[COMP_Y] + s1[COMP_Z]*s2[COMP_Z]);
}

// Vector - vector interpolation
inline	void	interpolatev(float *result,const float *s1,const float *s2,const float alpha) {
	result[0]	=	s1[0]*(1-alpha) + s2[0]*alpha;
	result[1]	=	s1[1]*(1-alpha) + s2[1]*alpha;
	result[2]	=	s1[2]*(1-alpha) + s2[2]*alpha;
}

// Move a vector
inline	void	movvv(float *dest,const float *src) {
	dest[0]		=	src[0];
	dest[1]		=	src[1];
	dest[2]		=	src[2];
}


// Length of a vector
inline	float	lengthv(const float *v) {
	return (float) sqrt(dotvv(v,v));
}

// Normalize a vector
inline	void	normalizev(float *v) {
	float	l	=	lengthv(v);

	v[0]		/=	l;
	v[1]		/=	l;
	v[2]		/=	l;
}

// Normalize a vector
inline	void	normalizev(float *r,float *v) {
	float	l	=	lengthv(v);

	r[0]	=	v[0]		/	l;
	r[1]	=	v[1]		/	l;
	r[2]	=	v[2]		/	l;
}


// Matrix - matrix multiplication (result should be a pointer to 16 floats)
inline	void	mulmm(float *result,const float *s1,const float *s2) {
	int i,j,k;

	for (i=0;i<4;i++)
		for (j=0;j<4;j++) {
				result[element(i,j)] = 0;
				for (k=0;k<4;k++) result[element(i,j)]+=s1[element(i,k)]*s2[element(k,j)];
		}
}

// Matrix - vector multiplication (result should be a pointer to 3 floats)
inline	void	mulmv(float *result,const float *s1,const float *s2) {
	float	w		= 	s1[element(3,0)]*s2[0]+s1[element(3,1)]*s2[1]+s1[element(3,2)]*s2[2]+s1[element(3,3)];

	result[0]	=	s1[element(0,0)]*s2[0]+s1[element(0,1)]*s2[1]+s1[element(0,2)]*s2[2]+s1[element(0,3)];
	result[1]	=	s1[element(1,0)]*s2[0]+s1[element(1,1)]*s2[1]+s1[element(1,2)]*s2[2]+s1[element(1,3)];
	result[2]	=	s1[element(2,0)]*s2[0]+s1[element(2,1)]*s2[1]+s1[element(2,2)]*s2[2]+s1[element(2,3)];
	

	if (w != 1) {	// FIXME: I may want to be more lenient here
		result[0] /= w;
		result[1] /= w;
		result[2] /= w;
	}
}

// Matrix - hvector multiplication (result should be a pointer to 4 floats)
inline	void	mulmhv(float *result,const float *s1,const float *s2) {
	result[0]	=	s1[element(0,0)]*s2[0]+s1[element(0,1)]*s2[1]+s1[element(0,2)]*s2[2]+s1[element(0,3)]*s2[3];
	result[1]	=	s1[element(1,0)]*s2[0]+s1[element(1,1)]*s2[1]+s1[element(1,2)]*s2[2]+s1[element(1,3)]*s2[3];
	result[2]	=	s1[element(2,0)]*s2[0]+s1[element(2,1)]*s2[1]+s1[element(2,2)]*s2[2]+s1[element(2,3)]*s2[3];
	result[3]	=	s1[element(3,0)]*s2[0]+s1[element(3,1)]*s2[1]+s1[element(3,2)]*s2[2]+s1[element(3,3)]*s2[3];
}

// Vector - matrix multiplication (result should be a pointer to 3 floats)
inline	void	mulvm(float  *result,const float *s1,const float *s2) {
	float	w		=	s1[0]*s2[element(0,3)] + s1[1]*s2[element(1,3)] + s1[2]*s2[element(2,3)] + s2[element(3,3)];
	
	result[0]	=	s1[0]*s2[element(0,0)] + s1[1]*s2[element(1,0)] + s1[2]*s2[element(2,0)] + s2[element(3,0)];
	result[1]	=	s1[0]*s2[element(0,1)] + s1[1]*s2[element(1,1)] + s1[2]*s2[element(2,1)] + s2[element(3,1)];
	result[2]	=	s1[0]*s2[element(0,2)] + s1[1]*s2[element(1,2)] + s1[2]*s2[element(2,2)] + s2[element(3,2)];


	if (w != 1) {	// FIXME: I may want to be more lenient here
		result[0] /= w;
		result[1] /= w;
		result[2] /= w;
	}
}

// Matrix - scalar multiplication
inline	void	mulmf(float  *result,const float *s,const float t) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	=	s[i]*t;
}

// Matrix - scalar multiplication
inline	void	mulmf(float  *result,const float t) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	*=	t;
}

// Matrix - matrix addition
inline	void	addmm(float  *result,const float *s1,const float *s2) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	=	s1[i] + s2[i];
}

// Matrix - matrix subtraction
inline	void	submm(float  *result,const float *s1,const float *s2) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	=	s1[i] - s2[i];
}

// Matrix - matrix addition
inline	void	addmm(float  *result,const float *s) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	+=	s[i];
}

// Matrix - matrix subtraction
inline	void	submm(float  *result,const float *s) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	-=	s[i];
}


// Move a matrix
inline	void	movmm(float *dest,const float *src) {
	int i;

	for (i=0;i<16;i++)
		dest[i]	=	src[i];
}

inline void		transposem(float *dest,const float *src) {
	int i,j;

	for (i=0;i<4;i++) {
		for (j=0;j<4;j++) {
			dest[element(i,j)] = src[element(j,i)];
		}
	}
}

inline float	area(float x1,float y1,float x2,float y2,float x3,float y3) {
	return (x1*y2 + y1*x3 + x2*y3 - x1*y3 - y1*x2 - y2*x3);
}


#ifdef USE_SSE
#include "xmmintrin.h"
#include "fvec.h"

typedef	F32vec4	ssevector[3];
typedef	F32vec4	ssematrix[16];


// Initialization functions
inline	void	initv(F32vec4 *r,const F32vec4 &x,const F32vec4 &y,const F32vec4 &z) {
	r[COMP_X]	=	x;
	r[COMP_Y]	=	y;
	r[COMP_Z]	=	z;
}

inline	void	initc(F32vec4 *r,const F32vec4 &red,const F32vec4 &green,const F32vec4 &blue) {
	r[COMP_R]	=	red;
	r[COMP_G]	=	green;
	r[COMP_B]	=	blue;
}

// Vector - vector addition
inline	void	addvv(F32vec4 *result,const F32vec4 *s1,const F32vec4 *s2) {
	result[0]	=	s1[0] + s2[0];
	result[1]	=	s1[1] + s2[1];
	result[2]	=	s1[2] + s2[2];
}

inline	void	addvv(F32vec4 *result,const F32vec4 *s1) {
	result[0]	=	result[0] + s1[0];
	result[1]	=	result[1] + s1[1];
	result[2]	=	result[2] + s1[2];
}

// Vector - vector subtraction
inline	void	subvv(F32vec4 *result,const F32vec4 *s1,const F32vec4 *s2) {
	result[0]	=	s1[0] - s2[0];
	result[1]	=	s1[1] - s2[1];
	result[2]	=	s1[2] - s2[2];
}

inline	void	subvv(F32vec4 *result,const F32vec4 *s1) {
	result[0]	=	result[0] - s1[0];
	result[1]	=	result[1] - s1[1];
	result[2]	=	result[2] - s1[2];
}

// Vector - float multiplication
inline	void	mulvf(F32vec4 *result,const F32vec4 *v,const float m) {
	F32vec4	qm(m);

	result[0]	=	v[0]*qm;
	result[1]	=	v[1]*qm;
	result[2]	=	v[2]*qm;
}

inline	void	mulvf(F32vec4 *result,const float m) {
	F32vec4	qm(m);

	result[0]	=	result[0]*qm;
	result[1]	=	result[1]*qm;
	result[2]	=	result[2]*qm;
}

// Vector - float multiplication
inline	void	mulvf(F32vec4 *result,const F32vec4 *v,const F32vec4 &m) {
	result[0]	=	v[0]*m;
	result[1]	=	v[1]*m;
	result[2]	=	v[2]*m;
}

inline	void	mulvf(F32vec4 *result,const F32vec4 &m) {
	result[0]	=	result[0]*m;
	result[1]	=	result[1]*m;
	result[2]	=	result[2]*m;
}


// Vector - vector multiplication (component by component)
inline	void	mulvv(F32vec4	*result,const F32vec4 *s1,const F32vec4 *s2) {
	result[0]	=	s1[0]*s2[0];
	result[1]	=	s1[1]*s2[1];
	result[2]	=	s1[2]*s2[2];
}

inline	void	mulvv(F32vec4 *result,const F32vec4 *s1) {
	result[0]	=	result[0]*s1[0];
	result[1]	=	result[1]*s1[1];
	result[2]	=	result[2]*s1[2];
}

// Vector - vector division (component by component)
inline	void	divvv(F32vec4 *result,const F32vec4 *s1,const F32vec4 *s2) {
	result[0]	=	s1[0] / s2[0];
	result[1]	=	s1[1] / s2[1];
	result[2]	=	s1[2] / s2[2];
}

inline	void	divvv(F32vec4 *result,const F32vec4 *s1) {
	result[0]	=	result[0] / s1[0];
	result[1]	=	result[1] / s1[1];
	result[2]	=	result[2] / s1[2];
}


// Vector - vector cross product
inline	void	crossvv(F32vec4 *result,const F32vec4 *s1,const F32vec4 *s2) {
	result[COMP_X]	=	s1[COMP_Y]*s2[COMP_Z] - s1[COMP_Z]*s2[COMP_Y];
	result[COMP_Y]	=	s1[COMP_Z]*s2[COMP_X] - s1[COMP_X]*s2[COMP_Z];
	result[COMP_Z]	=	s1[COMP_X]*s2[COMP_Y] - s1[COMP_Y]*s2[COMP_X];
}

// Vector - vector dot product
inline	F32vec4	dotvv(const F32vec4 *s1,const F32vec4 *s2) {
	return s1[COMP_X]*s2[COMP_X]+s1[COMP_Y]*s2[COMP_Y]+s1[COMP_Z]*s2[COMP_Z];
}


// Vector - vector interpolation
inline	void	interpolatev(F32vec4 *result,const F32vec4 *s1,const F32vec4 *s2,const float alpha) {
	F32vec4	oneMinusAlpha(1-alpha);
	F32vec4	Alpha(alpha);
	result[0]	=	s1[0]*oneMinusAlpha-s2[0]*Alpha;
	result[1]	=	s1[1]*oneMinusAlpha-s2[1]*Alpha;
	result[2]	=	s1[2]*oneMinusAlpha-s2[2]*Alpha;
}

// Move a vector
inline	void	movvv(F32vec4 *dest,const F32vec4 *src) {
	dest[0]		=	src[0];
	dest[1]		=	src[1];
	dest[2]		=	src[2];
}

inline	void	movvv(F32vec4 *dest,const float *src) {
	dest[0]		=	F32vec4(src[0]);
	dest[1]		=	F32vec4(src[1]);
	dest[2]		=	F32vec4(src[2]);
}

// Length of a vector
inline	F32vec4	lengthv(const F32vec4 *v) {
	return sqrt(dotvv(v,v));
}

// Normalize a vector
inline	void	normalizev(F32vec4 *v) {
	F32vec4	l	=	lengthv(v);

	v[0]		=	v[0]/l;
	v[1]		=	v[1]/l;
	v[2]		=	v[2]/l;
}

// Normalize a vector
inline	void	normalizev(F32vec4 *r,F32vec4 *v) {
	F32vec4	l	=	lengthv(v);

	r[0]		=	v[0]/l;
	r[1]		=	v[1]/l;
	r[2]		=	v[2]/l;
}


// Matrix - matrix multiplication (result should be a pointer to 16 floats)
inline	void	mulmm(F32vec4 *result,const F32vec4 *s1,const F32vec4 *s2) {
	int i,j,k;

	for (i=0;i<4;i++)
		for (j=0;j<4;j++) {
				result[element(i,j)] = F32vec4((float) 0);
				for (k=0;k<4;k++) 
					result[element(i,j)]	=	result[element(i,j)] + s1[element(i,k)]*s2[element(k,j)];
		}
}

// Matrix - vector multiplication (result should be a pointer to 3 floats)
inline	void	mulmv(F32vec4 *result,const float *s1,const F32vec4 *s2) {
	F32vec4	w		= 	F32vec4(s1[element(3,0)])*s2[0]+F32vec4(s1[element(3,1)])*s2[1]+F32vec4(s1[element(3,2)])*s2[2]+F32vec4(s1[element(3,3)]);

	result[0]	=	F32vec4(s1[element(0,0)])*s2[0]+F32vec4(s1[element(0,1)])*s2[1]+F32vec4(s1[element(0,2)])*s2[2]+F32vec4(s1[element(0,3)]);
	result[1]	=	F32vec4(s1[element(1,0)])*s2[0]+F32vec4(s1[element(1,1)])*s2[1]+F32vec4(s1[element(1,2)])*s2[2]+F32vec4(s1[element(1,3)]);
	result[2]	=	F32vec4(s1[element(2,0)])*s2[0]+F32vec4(s1[element(2,1)])*s2[1]+F32vec4(s1[element(2,2)])*s2[2]+F32vec4(s1[element(2,3)]);
	
	result[0] /= w;
	result[1] /= w;
	result[2] /= w;
}

// Matrix - hvector multiplication (result should be a pointer to 3 floats)
inline	void	mulmhv(F32vec4 *result,const F32vec4 *s1,const F32vec4 *s2) {
	result[0]	=	F32vec4(s1[element(0,0)])*s2[0]+F32vec4(s1[element(0,1)])*s2[1]+F32vec4(s1[element(0,2)])*s2[2]+F32vec4(s1[element(0,3)])*s2[3];
	result[1]	=	F32vec4(s1[element(1,0)])*s2[0]+F32vec4(s1[element(1,1)])*s2[1]+F32vec4(s1[element(1,2)])*s2[2]+F32vec4(s1[element(1,3)])*s2[3];
	result[2]	=	F32vec4(s1[element(2,0)])*s2[0]+F32vec4(s1[element(2,1)])*s2[1]+F32vec4(s1[element(2,2)])*s2[2]+F32vec4(s1[element(2,3)])*s2[3];
	result[3]	=	F32vec4(s1[element(3,0)])*s2[0]+F32vec4(s1[element(3,1)])*s2[1]+F32vec4(s1[element(3,2)])*s2[2]+F32vec4(s1[element(3,3)])*s2[3];
}

// Vector - matrix multiplication (result should be a pointer to 3 floats)
inline	void	mulvm(F32vec4  *result,const F32vec4 *s1,const F32vec4 *s2) {
	F32vec4	w		=	s1[0]*F32vec4(s2[element(0,3)]) + s1[1]*F32vec4(s2[element(1,3)]) + s1[2]*F32vec4(s2[element(2,3)]) + F32vec4(s2[element(3,3)]);
	
	result[0]	=	s1[0]*F32vec4(s2[element(0,0)]) + s1[1]*F32vec4(s2[element(1,0)]) + s1[2]*F32vec4(s2[element(2,0)]) + F32vec4(s2[element(3,0)]);
	result[1]	=	s1[0]*F32vec4(s2[element(0,1)]) + s1[1]*F32vec4(s2[element(1,1)]) + s1[2]*F32vec4(s2[element(2,1)]) + F32vec4(s2[element(3,1)]);
	result[2]	=	s1[0]*F32vec4(s2[element(0,2)]) + s1[1]*F32vec4(s2[element(1,2)]) + s1[2]*F32vec4(s2[element(2,2)]) + F32vec4(s2[element(3,2)]);

	result[0] /= w;
	result[1] /= w;
	result[2] /= w;
}

// Matrix - scalar multiplication
inline	void	mulmf(F32vec4  *result,const F32vec4 *s,const float t) {
	F32vec4	T(t);
	int		i;

	for (i=0;i<16;i++)
		result[i]	=	s[i]*T;
}

// Matrix - scalar multiplication
inline	void	mulmf(F32vec4  *result,const float t) {
	F32vec4	T(t);
	int	i;

	for (i=0;i<16;i++)
		result[i]	=	result[i]*T;
}

// Matrix - matrix addition
inline	void	addmm(F32vec4  *result,const F32vec4 *s1,const F32vec4 *s2) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	=	s1[i] * s2[i];
}

// Matrix - matrix subtraction
inline	void	submm(F32vec4  *result,const F32vec4 *s1,const F32vec4 *s2) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	=	s1[i] - s2[i];
}

// Matrix - matrix addition
inline	void	addmm(F32vec4  *result,const F32vec4 *s) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	=	result[i] + s[i];
}

// Matrix - matrix subtraction
inline	void	submm(F32vec4  *result,const F32vec4 *s) {
	int	i;

	for (i=0;i<16;i++)
		result[i]	=	result[i] - s[i];
}


// Move a matrix
inline	void	movmm(F32vec4 *dest,const F32vec4 *src) {
	int i;

	for (i=0;i<16;i++)
		dest[i]	=	src[i];
}

// Transpose a matrix
inline void		transposem(F32vec4 *dest,const F32vec4 *src) {
	int i,j;

	for (i=0;i<4;i++) {
		for (j=0;j<4;j++) {
			dest[element(i,j)] = src[element(j,i)];
		}
	}
}

// Some comparisons (semantics : at least one item satisfies)
inline	int		operator==(const F32vec4 &s1,const F32vec4 &s2) {
	F32vec4	res		=	cmpeq(s1,s2);
	unsigned int		*resi	=	(unsigned int *) &res;

	return	(int) ((resi[0] >> 31) + (resi[1] >> 31) + (resi[2] >> 31) + (resi[3] >> 31));
}

inline	int		operator!=(const F32vec4 &s1,const F32vec4 &s2) {
	F32vec4	res		=	cmpneq(s1,s2);
	unsigned int		*resi	=	(unsigned int *) &res;

	return	(int) ((resi[0] >> 31) + (resi[1] >> 31) + (resi[2] >> 31) + (resi[3] >> 31));
}

inline	int		operator<(const F32vec4 &s1,const F32vec4 &s2) {
	F32vec4	res		=	cmplt(s1,s2);
	unsigned int		*resi	=	(unsigned int *) &res;

	return	(int) ((resi[0] >> 31) + (resi[1] >> 31) + (resi[2] >> 31) + (resi[3] >> 31));
}

inline	int		operator>(const F32vec4 &s1,const F32vec4 &s2) {
	F32vec4	res		=	cmpgt(s1,s2);
	unsigned int		*resi	=	(unsigned int *) &res;

	return	(int) ((resi[0] >> 31) + (resi[1] >> 31) + (resi[2] >> 31) + (resi[3] >> 31));
}

inline	int		operator<=(const F32vec4 &s1,const F32vec4 &s2) {
	F32vec4	res		=	cmple(s1,s2);
	unsigned int		*resi	=	(unsigned int *) &res;

	return	(int) ((resi[0] >> 31) + (resi[1] >> 31) + (resi[2] >> 31) + (resi[3] >> 31));
}

inline	int		operator>=(const F32vec4 &s1,const F32vec4 &s2) {
	F32vec4	res		=	cmpge(s1,s2);
	unsigned int		*resi	=	(unsigned int *) &res;

	return	(int) ((resi[0] >> 31) + (resi[1] >> 31) + (resi[2] >> 31) + (resi[3] >> 31));
}


#endif


inline	void	mulmvv(float *dest,const float *s1,const float *s2) {
	dest[element(0,0)]	=	s1[0]*s2[0];
	dest[element(1,0)]	=	s1[1]*s2[0];
	dest[element(2,0)]	=	s1[2]*s2[0];
	dest[element(3,0)]	=	0;
	dest[element(0,1)]	=	s1[0]*s2[1];
	dest[element(1,1)]	=	s1[1]*s2[1];
	dest[element(2,1)]	=	s1[2]*s2[1];
	dest[element(3,1)]	=	0;
	dest[element(0,2)]	=	s1[0]*s2[2];
	dest[element(1,2)]	=	s1[1]*s2[2];
	dest[element(2,2)]	=	s1[2]*s2[2];
	dest[element(3,2)]	=	0;
	dest[element(0,3)]	=	0;
	dest[element(1,3)]	=	0;
	dest[element(2,3)]	=	0;
	dest[element(3,3)]	=	1;
}

inline	float	clampf(const float mi,const float l,const float ma) {
	return (l < mi ? mi : (l > ma ? ma : l));
}

inline	void	clampv(const float *mi,float *l,const float *ma) {
	l[0]	=	(l[0] < mi[0] ? mi[0] : (l[0] > ma[0] ? ma[0] : l[0]));
	l[1]	=	(l[1] < mi[1] ? mi[1] : (l[1] > ma[1] ? ma[1] : l[1]));
	l[2]	=	(l[2] < mi[2] ? mi[2] : (l[2] > ma[2] ? ma[2] : l[2]));
}

inline	int		inBox(const float *bmin,const float *bmax,const float *v) {
	int	i;

	for (i=0;i<3;i++) {
		if ((v[i] < bmin[i]) || (v[i] > bmax[i])) return FALSE;
	}

	return TRUE;
}


inline	void	addBox(float *bmin,float *bmax,const float *v) {
	int	i;

	for (i=0;i<3;i++) {
		if (v[i] < bmin[i])	bmin[i]	=	v[i];
		if (v[i] > bmax[i]) bmax[i] =	v[i];
	}
}

inline	int		intersectBox(const float *bmin1,const float *bmax1,const float *bmin2,const float *bmax2) {
	int	i;

	for (i=0;i<3;i++) {
		if ((bmin1[i] > bmax2[i]) || (bmax1[i] < bmin2[i]))	return	FALSE;
	}

	return	TRUE;
}


#define     cbrt(x)     ((x) > 0.0 ? pow((T)(x), 1.0/3.0) : \
						((x) < 0.0 ? -pow((T)-(x), 1.0/3.0) : 0.0))

template <class T> inline	int	solveQuadric(T a,T b,T c,T *r) {
	T	inva	=	1	/	a;
	T	nc		=	c*inva;
	T	nb		=	b*inva*(T) 0.5;

	T	delta	=	nb*nb - nc;

	if (delta < 0)	return	0;
	else {
		T	sqrtDelta	=	(T) sqrt(delta);

		r[0]	=	-sqrtDelta - nb;
		r[1]	=	sqrtDelta - nb;

		return 2;
	}
}

template <class T> inline	int solveCubic(T c[4],T s[3])
{
    int     i, num;
    T		sub;
    T		A, B, C;
    T		sq_A, p, q;
    T		cb_p, D;

    /* normal form: x^3 + Ax^2 + Bx + C = 0 */

    A = c[ 2 ] / c[ 3 ];
    B = c[ 1 ] / c[ 3 ];
    C = c[ 0 ] / c[ 3 ];

    /*  substitute x = y - A/3 to eliminate quadric term:
	x^3 +px + q = 0 */

    sq_A	= A * A;
    p		= (T) (1.0/3 * (- 1.0/3 * sq_A + B));
    q		= (T) (1.0/2 * (2.0/27 * A * sq_A - 1.0/3 * A * B + C));

    /* use Cardano's formula */

    cb_p	= p * p * p;
    D		= q * q + cb_p;

    if (zero(D)) {
		if (zero(q)) {
			s[ 0 ] = 0;
			num = 1;
		} else 	{
			T u		= (T) (cbrt(-q));
			s[ 0 ]	= 2 * u;
			s[ 1 ]	= - u;
			num = 2;
		}
    } else if (D < 0) {
		T phi	= (T) (1.0/3 * acos(-q / sqrt(-cb_p)));
		T t		= (T) (2 * sqrt(-p));

		s[ 0 ] =   t * cos(phi);
		s[ 1 ] = - t * cos(phi + PI / 3);
		s[ 2 ] = - t * cos(phi - PI / 3);
		num = 3;
    } else {
		T sqrt_D	= (T) (sqrt(D));
		T u			= (T) (cbrt(sqrt_D - q));
		T v			= (T) (-cbrt(sqrt_D + q));

		s[ 0 ]		= u + v;
		num			= 1;
    }

    sub = (T) (1.0/3 * A);

    for (i = 0; i < num; ++i)
		s[ i ] -= sub;

    return num;
}


template <class T> inline	int solveQuartic(T c[5],T s[4]) {
    T		coeffs[ 4 ];
    T		z, u, v, sub;
    T		A, B, C, D;
    T		sq_A, p, q, r;
    int     i, num;

    A = c[ 3 ] / c[ 4 ];
    B = c[ 2 ] / c[ 4 ];
    C = c[ 1 ] / c[ 4 ];
    D = c[ 0 ] / c[ 4 ];


    sq_A	= A * A;
    p		= (T) (- 3.0/8 * sq_A + B);
    q		= (T) (1.0/8 * sq_A * A - 1.0/2 * A * B + C);
    r		= (T) (-3.0/256*sq_A*sq_A + 1.0/16*sq_A*B - 1.0/4*A*C + D);

    if (zero(r)) {
		coeffs[ 0 ] = q;
		coeffs[ 1 ] = p;
		coeffs[ 2 ] = 0;
		coeffs[ 3 ] = 1;

		num = solveCubic<T>(coeffs, s);

		s[ num++ ] = 0;
    } else {
		coeffs[ 0 ] = 1.0/2 * r * p - 1.0/8 * q * q;
		coeffs[ 1 ] = - r;
		coeffs[ 2 ] = - 1.0/2 * p;
		coeffs[ 3 ] = 1;

		(void) solveCubic<T>(coeffs, s);

		z = s[ 0 ];

		u = z * z - r;
		v = 2 * z - p;

		if (zero(u))
			u = 0;
		else if (u > 0)
			u = sqrt(u);
		else
			return 0;

		if (zero(v))
			v = 0;
		else if (v > 0)
			v = sqrt(v);
		else
			return 0;

		coeffs[ 0 ] = z - u;
		coeffs[ 1 ] = q < 0 ? -v : v;
		coeffs[ 2 ] = 1;

		num = solveQuadric<T>(coeffs[2],coeffs[1],coeffs[0],s);

		coeffs[ 0 ]= z + u;
		coeffs[ 1 ] = q < 0 ? v : -v;
		coeffs[ 2 ] = 1;

		num += solveQuadric<T>(coeffs[2],coeffs[1],coeffs[0],s + num);
    }

    /* resubstitute */

    sub = 1.0/4 * A;

    for (i = 0; i < num; ++i)
	s[ i ] -= sub;

    return num;
}

// Reflect I around N
inline	void	reflect(float *r,const float *N,const float *I) {
	mulvf(r,N,-2*dotvv(N,I));
	addvv(r,I);
}

// Refrect I around N
inline	void	refract(float *r,const float *N,const float *I,float eta) {
	vector	vtmp;
	float	IdotN,k;
	IdotN	=	dotvv(N,I);
	k		=	1 - eta*eta*(1-IdotN*IdotN);
	if (k < 0)	initv(r,0,0,0);
	else	{
		mulvf(vtmp,N,(eta*IdotN+(float) sqrt(k)));
		mulvf(r,I,eta);
		subvv(r,vtmp);
	}	
}

// Unit random number generator
inline	float	urand() {
	return	rand() / (float) RAND_MAX;
}

void	exponentmv(float *,const float *);								// Constructs a rotation matrix from exponential rep
void	exponentvm(float *,const float *);								// Constructs a rotation matrix from exponential rep
void	skewsymm(float *,const float *);								// Create a skew symmetric matrix from matrix
float	determinantm(const float *);									// Take the determinant of a 4x4 matrix
void	identitym(float *);												// Construct identity matrix
void	translatem(float *,const float,const float,const float);		// Construct translate matrix
void	scalem(float *,const float,const float,const float);			// Construct scale matrix
void	rotatem(float *,const float *,const float);						// Construct rotate matrix (the second argument is the rotation vector)
void	rotatem(float *,float,float,float,const float);					// Construct rotate matrix (the arguments are: result,x,y,z coordinates of the rotation vector and angle to rotate)
void	rotatem(float *,const float *);
void	skewm(float *,const float,const float,const float,const float,const float,const float,const float);
int		invertm(float *,const float *);									// Invert a matrix. Returns 0 if the inversion is successful 1 otherwise (i.e. matrix is singular)
int		dsvd(float **a, int m, int n, float *w, float **v);				// SVD

#endif


