///////////////////////////////////////////////////////////////////////
//
//  File				:	simulation.h
//  Classes				:	CGeometry
//							CForce
//							CRigidBody
//							CCollisionDetector
//							CSimulation
//  Description			:	This file defines the simulator interface
//
//
//
//
//
////////////////////////////////////////////////////////////////////////
#ifndef SIMULATION_H
#define SIMULATION_H

#include	"common/algebra.h"
#include	"common/containers.h"
#include	"psystem.h"
#include	<SWIFT.h>

// If this flag is set, the body does not move
const	int		RBODY_FIXED			=	1;

///////////////////////////////////////////////////////////////////////
// Class				:	CGeometry
// Description			:	This class defines a geometry that will
//							be used during the collisition detection
// Comments				:	Currently only convex geometries are supported
// Date last edited		:	4/15/2001
class	CGeometry {
public:
	class	CTriangle {
	public:
		int			v[3];			// Pointers to the vertices
		vector		norm;			// The normal vector
		float		d;				// The distance along normal to the origin

		void		computeNorm(vector *);
	};

public:
				CGeometry();
	virtual		~CGeometry();

	void		finalize();

	int			numVertices;
	int			numTriangles;
	vector		*vertices;			// The positions only
	CTriangle	*triangles;




	int			collidingVertex;
	int			collidingTriangle;
};

///////////////////////////////////////////////////////////////////////
// Class				:	CCollisionDetector
// Description			:	Obvious
// Comments				:
// Date last edited		:	4/15/2001
class	CRigidBody;

const	unsigned	int	COLLISION_DETECT_SINGLE	=	1;

const	unsigned	int	COLLISION_CLEAR			=	0;
const	unsigned	int	COLLISION_HANDLED		=	1;
const	unsigned	int	COLLISION_SUBDIVIDE		=	2;

class	CCollisionDetector {
	class	CContact {
	public:
		vector		normal;
		vector		P;
		CRigidBody	*body1,*body2;
		vector		P1,P2;

		////////////////////////////////////////////////////////////////
		vector		vel1,vel2,vrel;
		float		ndotv;
	};

public:
					CCollisionDetector();
					~CCollisionDetector();

	virtual		void	*registerGeometry(CRigidBody *,CArray<CGeometry *> *,int fixed = FALSE);
	virtual		void	setState(void *,float *,float *);
	virtual		int		check();

				void	computeVelocities(CContact *);
				void	handleContact(CContact *);
				void	frictionlessHandler(CContact *,float);

				float	tolerance;		// Penetration tolerance
				int		flags;

	CContact			*contacts;
	int					nContacts,mContacts;

	float				**coefficientsOfFriction;
	float				**coefficientsOfRestitution;
	int					numBodies,maxBodies;
	float				defaultFriction,defaultRestitution;
};

///////////////////////////////////////////////////////////////////////
// Class				:	CSWIFT
// Description			:	SWIFT based collision detector
// Comments				:
// Date last edited		:	4/15/2001
class	CSWIFT : public CCollisionDetector {
public:
				CSWIFT();
				~CSWIFT();

			void	*registerGeometry(CRigidBody *,CArray<CGeometry *> *,int fixed = FALSE);
			void	setState(void *,float *,float *);
			int		check();
private:
	SWIFT_Scene				*scene;
	int						*objectPairs;
	int						*featureTypes;
	int						*featureIds;
	int						*pointIds;
	SWIFT_Real				*normals;
	SWIFT_Real				*distances;
	SWIFT_Real				*contactPoints;
	CArray<CRigidBody *>	bodies;
};

///////////////////////////////////////////////////////////////////////
// Class				:	CRigidBody
// Description			:	Obvious
// Comments				:
// Date last edited		:	4/15/2001
class	CRigidBody {
public:
	///////////////////////////////////////////////////////////////////////
	// Class				:	CState
	// Description			:	This class holds all the state variables
	// Comments				:
	// Date last edited		:	4/15/2001
	class	CState {
	public:
		vector		position;			// The position (of COM in world coordinates)
		float		orientation[4];		// Orientation (quaternion)
		vector		P;					// Linear momentum
		vector		L;					// Angular momentum

		vector		velocity;			// Linear velocity
		vector		avelocity;			// Angular velocity

		matrix		R,Rt;				// The rotation matrix corresponding to the orientation
		matrix		Iinv;

		float		clock;				// The time of this state
	};

public:
				CRigidBody(CCollisionDetector *,CArray<CGeometry *> *,float);
	virtual		~CRigidBody();

	void		advance(float dTime);
	void		rollback();
	void		commit();

	CState		*currentState;			// This state is used temporarily to keep the latest state
	CState		*lastStableState;		// This holds the last stable state

	vector		COM;					// Center Of Mass (in object coordinates)
	float		mass;					// The total mass of the body
	float		invMass;				// The inverse of the body mass

	int			flags;					// Holds misc switches

										// The states for this rigid body
	CState		states[2];
										// Precomputed inertia tensor
	matrix		inertiaTensor,invInertiaTensor;
										// The handle of the body given by the collisiton detector
	void		*collisionHandle;
										// The geometry of the rigid body
	CArray<CGeometry *> *geometry;

	vector		netForce;
	vector		netTorque;

	int			bodyIndex;
};


///////////////////////////////////////////////////////////////////////
// Class				:	CSimulation
// Description			:	Obvious
// Comments				:
// Date last edited		:	4/15/2001
class	CSimulation {
public:
	class	CBodyEntry {
	public:
		CRigidBody			*body;
		CRigidBody::CState	initialState;
	};

public:
							CSimulation(CCollisionDetector *customDetector = NULL);
	virtual					~CSimulation();

	void					addPsystem(CParticleSystem *);
	void					addObject(CRigidBody *);
	virtual	void			addForces(CRigidBody *);	// Celled by the simulator to accumulate forces on the object
	void					simulate(float);
	void					reset();
	virtual	void			geometry(CGeometry *,float *,float *)	{	}
	virtual	void			step(float)								{	}

	float					clock;						// The simulation clock

	int						numBodies,maxBodies;		// The bodies array
	CBodyEntry				*bodies;

	vector					gravity;					// Preset gravity

	CCollisionDetector		*collisionDetector;			// The collision detector

	CParticleSystem			*psystems;					// The particle systems

	float					maxTimeStep;
	float					velocityDamping,avelocityDamping;

private:
	void					advance(float);
};

extern	void	computeInertiaTensor(float *,float *,float,CArray<CGeometry *> *);

#endif
