#include "simulation.h"

#define SQR(x) ((x)*(x))
#define CUBE(x) ((x)*(x)*(x))

#define X 0
#define Y 1
#define Z 2


/*
   ============================================================================
   globals
   ============================================================================
*/


static int A;   /* alpha */
static int B;   /* beta */
static int C;   /* gamma */

/* projection integrals */
static double P1, Pa, Pb, Paa, Pab, Pbb, Paaa, Paab, Pabb, Pbbb;

/* face integrals */
static double Fa, Fb, Fc, Faa, Fbb, Fcc, Faaa, Fbbb, Fccc, Faab, Fbbc, Fcca;

/* volume integrals */
static double T0, T1[3], T2[3], TP[3];



void compProjectionIntegrals(CGeometry *geo,CGeometry::CTriangle *f)
{
  double a0, a1, da;
  double b0, b1, db;
  double a0_2, a0_3, a0_4, b0_2, b0_3, b0_4;
  double a1_2, a1_3, b1_2, b1_3;
  double C1, Ca, Caa, Caaa, Cb, Cbb, Cbbb;
  double Cab, Kab, Caab, Kaab, Cabb, Kabb;
  int i;

  P1 = Pa = Pb = Paa = Pab = Pbb = Paaa = Paab = Pabb = Pbbb = 0.0;

  for (i = 0; i < 3; i++) {
    a0 = geo->vertices[f->v[i]][A];;
    b0 = geo->vertices[f->v[i]][B];
    a1 = geo->vertices[f->v[(i+1) % 3]][A];
    b1 = geo->vertices[f->v[(i+1) % 3]][B];
    da = a1 - a0;
    db = b1 - b0;
    a0_2 = a0 * a0; a0_3 = a0_2 * a0; a0_4 = a0_3 * a0;
    b0_2 = b0 * b0; b0_3 = b0_2 * b0; b0_4 = b0_3 * b0;
    a1_2 = a1 * a1; a1_3 = a1_2 * a1; 
    b1_2 = b1 * b1; b1_3 = b1_2 * b1;

    C1 = a1 + a0;
    Ca = a1*C1 + a0_2; Caa = a1*Ca + a0_3; Caaa = a1*Caa + a0_4;
    Cb = b1*(b1 + b0) + b0_2; Cbb = b1*Cb + b0_3; Cbbb = b1*Cbb + b0_4;
    Cab = 3*a1_2 + 2*a1*a0 + a0_2; Kab = a1_2 + 2*a1*a0 + 3*a0_2;
    Caab = a0*Cab + 4*a1_3; Kaab = a1*Kab + 4*a0_3;
    Cabb = 4*b1_3 + 3*b1_2*b0 + 2*b1*b0_2 + b0_3;
    Kabb = b1_3 + 2*b1_2*b0 + 3*b1*b0_2 + 4*b0_3;

    P1 += db*C1;
    Pa += db*Ca;
    Paa += db*Caa;
    Paaa += db*Caaa;
    Pb += da*Cb;
    Pbb += da*Cbb;
    Pbbb += da*Cbbb;
    Pab += db*(b1*Cab + b0*Kab);
    Paab += db*(b1*Caab + b0*Kaab);
    Pabb += da*(a1*Cabb + a0*Kabb);
  }

  P1 /= 2.0;
  Pa /= 6.0;
  Paa /= 12.0;
  Paaa /= 20.0;
  Pb /= -6.0;
  Pbb /= -12.0;
  Pbbb /= -20.0;
  Pab /= 24.0;
  Paab /= 60.0;
  Pabb /= -60.0;
}

void compFaceIntegrals(CGeometry *geo,CGeometry::CTriangle *f)
{
  double w;
  double k1, k2, k3, k4;
  double n[3];

  compProjectionIntegrals(geo,f);

  w		= f->d;
  n[0]	=	f->norm[0];
  n[1]	=	f->norm[1];
  n[2]	=	f->norm[2];

  k1 = 1 / n[C]; k2 = k1 * k1; k3 = k2 * k1; k4 = k3 * k1;

  Fa = k1 * Pa;
  Fb = k1 * Pb;
  Fc = -k2 * (n[A]*Pa + n[B]*Pb + w*P1);

  Faa = k1 * Paa;
  Fbb = k1 * Pbb;
  Fcc = k3 * (SQR(n[A])*Paa + 2*n[A]*n[B]*Pab + SQR(n[B])*Pbb
	 + w*(2*(n[A]*Pa + n[B]*Pb) + w*P1));

  Faaa = k1 * Paaa;
  Fbbb = k1 * Pbbb;
  Fccc = -k4 * (CUBE(n[A])*Paaa + 3*SQR(n[A])*n[B]*Paab 
	   + 3*n[A]*SQR(n[B])*Pabb + CUBE(n[B])*Pbbb
	   + 3*w*(SQR(n[A])*Paa + 2*n[A]*n[B]*Pab + SQR(n[B])*Pbb)
	   + w*w*(3*(n[A]*Pa + n[B]*Pb) + w*P1));

  Faab = k1 * Paab;
  Fbbc = -k2 * (n[A]*Pabb + n[B]*Pbbb + w*Pbb);
  Fcca = k3 * (SQR(n[A])*Paaa + 2*n[A]*n[B]*Paab + SQR(n[B])*Pabb
	 + w*(2*(n[A]*Paa + n[B]*Pab) + w*Pa));
}


void compVolumeIntegrals(CGeometry *p) {
  CGeometry::CTriangle *f;
  double nx, ny, nz;
  int i;

  for (i = 0; i < p->numTriangles; i++) {

    f = &p->triangles[i];

    nx = fabs(f->norm[X]);
    ny = fabs(f->norm[Y]);
    nz = fabs(f->norm[Z]);
    if (nx > ny && nx > nz) C = X;
    else C = (ny > nz) ? Y : Z;
    A = (C + 1) % 3;
    B = (A + 1) % 3;

    compFaceIntegrals(p,f);

    T0 += f->norm[X] * ((A == X) ? Fa : ((B == X) ? Fb : Fc));

    T1[A] += f->norm[A] * Faa;
    T1[B] += f->norm[B] * Fbb;
    T1[C] += f->norm[C] * Fcc;
    T2[A] += f->norm[A] * Faaa;
    T2[B] += f->norm[B] * Fbbb;
    T2[C] += f->norm[C] * Fccc;
    TP[A] += f->norm[A] * Faab;
    TP[B] += f->norm[B] * Fbbc;
    TP[C] += f->norm[C] * Fcca;
  }
}

void compVolumeIntegrals(CArray<CGeometry *> *p)
{
  int i;

  T0 = T1[X] = T1[Y] = T1[Z] 
     = T2[X] = T2[Y] = T2[Z] 
     = TP[X] = TP[Y] = TP[Z] = 0;

  for (i = 0; i < p->getSize(); i++) {
	  compVolumeIntegrals((*p)[i]);
  }

  T1[X] /= 2; T1[Y] /= 2; T1[Z] /= 2;
  T2[X] /= 3; T2[Y] /= 3; T2[Z] /= 3;
  TP[X] /= 2; TP[Y] /= 2; TP[Z] /= 2;
}


void	computeInertiaTensor(float *COM,float *inertia,float mass,CArray<CGeometry *> *geometry) {
  double density;

  identitym(inertia);

  compVolumeIntegrals(geometry);
	
  density	=	mass/T0;

  /* compute center of mass */
  COM[X] = (float) (T1[X] / T0);
  COM[Y] = (float) (T1[Y] / T0);
  COM[Z] = (float) (T1[Z] / T0);

  /* compute inertia tensor */
  inertia[element(0,0)]		= (float) (density * (T2[Y] + T2[Z]));
  inertia[element(1,1)]		= (float) (density * (T2[Z] + T2[X]));
  inertia[element(2,2)]		= (float) (density * (T2[X] + T2[Y]));
  inertia[element(0,1)]		= inertia[element(1,0)]	= (float) (-density * TP[X]);
  inertia[element(1,2)]		= inertia[element(2,1)]	= (float) (-density * TP[Y]);
  inertia[element(0,2)]		= inertia[element(2,0)]	= (float) (-density * TP[Z]);

  /* translate inertia tensor to center of mass */
  inertia[element(0,0)]		-= (float) (mass * (COM[Y]*COM[Y] + COM[Z]*COM[Z]));
  inertia[element(1,1)]		-= (float) (mass * (COM[Z]*COM[Z] + COM[X]*COM[X]));
  inertia[element(2,2)]		-= (float) (mass * (COM[X]*COM[X] + COM[Y]*COM[Y]));
  inertia[element(1,0)]		= inertia[element(0,1)] += (float) (mass * COM[X] * COM[Y]); 
  inertia[element(1,2)]		= inertia[element(2,1)] += (float) (mass * COM[Y] * COM[Z]); 
  inertia[element(0,2)]		= inertia[element(2,0)] += (float) (mass * COM[Z] * COM[X]); 


}